/** 
 * @file rexpx6x70d_dio.c
 * @brief DIO入出力ライブラリ
 * @version 1.0.0
 * @date 2012-02-17
 * @par Copyright
 *  (C) 2012 RATOC Systems, Inc.
 * @par History
 *  - 2012.02.17 v1.0.0
 *    -# Initial Version.
 */
/*
 *  Copyright (C) 2012 RATOC Systems, Inc.
 *  Dual licensed under the BSD (BSD-LICENSE.txt)
 *  and GPL (GPL-LICENSE.txt) licenses.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <semaphore.h>
#include <pthread.h>
#include <sys/ioctl.h>
#include <time.h>

#include "../driver/rexpx6x70d_drv.h"
#include "rexpx6x70d_dio.h"
#include "def_xr17.h"

/* ------------------------ */
/*	DEFINE					*/
/* ------------------------ */
//#define _DEBUG
#ifdef _DEBUG
#define DBG_PRINTF			printf
#define DBG_FPRINTF			fprintf
#else
#define DBG_PRINTF(...)		do { } while(0)
#define DBG_FPRINTF(...)	do { } while(0)
#endif /* _DEBUG */

/** 
 * @name デバイス
 * @{ */
#define MAX_DEVICE_ID			4										/**< デバイスID最大数								*/
#define MAX_DEVICE_NUM			(MAX_DEVICE_ID * eBOARD_KIND_MAX)		/**< デバイス最大数									*/
#define DIO_DEV_NAME_PCI		"/dev/rexdioPCI%d"						/**< デバイス名										*/
#define DIO_DEV_NAME_PE			"/dev/rexdioPE%d"						/**< デバイス名										*/
/**  @} */

/* ------------------------- */
/*	STRUCT^  ^   ^   ^   ^   */
/* ------------------------- */

/** 
 * @struct T_INTERRUPT_ARG
 * @brief 割り込みスレッドの引数
 */
typedef struct interrupt_ctrl {
	int					fd;												/**< file descriptor								*/
	PAPIFUNC			pFunc;											/**< コールバック関数								*/
	pthread_t			thread_id;										/**< スレッドID										*/
	int					int_start;										/**< 割り込み動作中フラグ							*/
	sem_t				semIntComplete;									/**< 割り込みスレッドから通知用						*/
} interrupt_ctrl_t, *pinterrupt_ctrl_t;									/**< 割り込みスレッド管理用変数の構造体				*/

/* ------------------------ */
/*	PRIVATE PROTTYPE		*/
/* ------------------------ */
static int CheckDirection(int fd, unsigned char *pDirection);
static int SetIntEdge(int fd, unsigned char IntEdge);
static unsigned char GetBoardID(unsigned char dvid, unsigned char value);
static void *InterruptThread(void *arg);
static int interrupt_args_add(int fd, int id, int kind);
static int interrupt_args_clear(int fd);
static pinterrupt_ctrl_t interrupt_args_get(int fd);

/* ------------------------ */
/*	GLOBAL VARIABLE			*/
/* ------------------------ */
/** 
 * @var g_interrupt_args
 * @brief 割り込み待ちスレッドの引数
 *
 * １デバイスにつき１スレッドまで動作させることが可能。
 *
 */
interrupt_ctrl_t g_interrupt_args[MAX_DEVICE_NUM] = {
	{-1, NULL, 0, 0},
	{-1, NULL, 0, 0},
	{-1, NULL, 0, 0},
	{-1, NULL, 0, 0},
	{-1, NULL, 0, 0},
	{-1, NULL, 0, 0},
	{-1, NULL, 0, 0},
	{-1, NULL, 0, 0},
};

/** 
 * @var g_sem_ref_interrupt_arg
 * @brief g_interrupt_argsのR/W時の排他用。
 */
sem_t g_sem_ref_interrupt_arg;											/**< 排他用セマフォ									*/

/** 
 * @var g_interrupt_args_init
 * @brief g_sem_ref_interrupt_argの初期化制御管理変数。
 */
int g_interrupt_args_init = 0;											/**< 初期化済みのチェック用							*/


/* ------------------------ */
/*	PUBLIC FUNCTION			*/
/* ------------------------ */
/** 
 * @brief ドライバーオープン
 * 
 * @param[in] BoardID	ボードID番号
 * @param[in] BoardKind	ボード種別
 * 
 * @retval 0以上				success. file descriptor value.
 * @retval eLIB_ERR_DRIVER		driver call error.
 * @retval eLIB_ERR_PARAM		parameter error.
 */
int DioOpen(int BoardID, board_kind_e BoardKind)
{
	char pathName[32];
	int hDriver;

	DBG_PRINTF("[%s(%d)] IN BoardID = %d, BoardKind = %d\n", __func__, __LINE__, BoardID, BoardKind);

	if(0 == g_interrupt_args_init)
	{
		g_interrupt_args_init = 1;
		sem_init(&g_sem_ref_interrupt_arg, 0, 1);
	}

	if(	(0 > BoardID)
	||	(MAX_DEVICE_ID <= BoardID)
	||	(eBOARD_KIND_MAX <= BoardKind))
	{
		return eLIB_ERR_PARAM;
	}

	memset(pathName, 0x00, sizeof(pathName));
	if(eBOARD_KIND_PE == BoardKind)
	{
		sprintf(pathName, DIO_DEV_NAME_PE, BoardID);
	}
	else
	{
		sprintf(pathName, DIO_DEV_NAME_PCI, BoardID);
	}
	hDriver = open(pathName, O_RDWR);

	if(0 <= hDriver)
	{
		/*	割り込みスレッド管理用変数に登録	*/
		interrupt_args_add(hDriver, BoardID, BoardKind);
	}
	else
	{
		fprintf(stderr, "open error : %s <%s>\n", pathName, strerror(errno));
		hDriver = eLIB_ERR_DRIVER;
	}

	DBG_PRINTF("[%s(%d)] OU hDriver = %d\n", __func__, __LINE__, hDriver);

	return hDriver;
}

/** 
 * @brief ドライバークローズ
 * 
 * @param[in] fd	file descriptor.
 * 
 * @retval 0				success.
 * @retval eLIB_ERR_DRIVER	driver call error.
 */
int DioClose(int fd)
{
	int nRet;

	DBG_PRINTF("[%s(%d)] IN fd = %d\n", __func__, __LINE__, fd);

	/* 割り込み待ちが動作している場合は止める */
	InterruptStop(fd);

	/*	割り込みスレッド管理用変数を初期化	*/
	interrupt_args_clear(fd);

	nRet = close(fd);
	if(0 > nRet)
	{
		fprintf(stderr, "close error : <%s>\n", strerror(errno));
		nRet = eLIB_ERR_DRIVER;
	}

	DBG_PRINTF("[%s(%d)] OU nRet = %d\n", __func__, __LINE__, nRet);

	return nRet;
}

/** 
 * @brief データ出力方向の設定
 * 
 * @param[in] fd			file descriptor.
 * @param[in] Direction		入力方向
 * 
 * @retval 0					success.
 * @retval eLIB_ERR_DRIVER		driver call error.
 *
 * 方向設定のためにMPIOSELのレジスタに値を設定する。@n
 * 入力方向に設定する場合は、対応するビットに1を設定する。
 */
int SetDirection(int fd, unsigned char Direction)
{
	int nRet;
	IOPORT_UCHAR bIOPort;

	bIOPort.Offset = MPIOSEL;
	bIOPort.Value = Direction;
	nRet = ioctl(fd, IOCTL_DIO_WRITE_PORT, &bIOPort);
	if(0 > nRet)
	{
		fprintf(stderr, "ioctl IOCTL_DIO_WRITE_PORT error : <%s>\n", strerror(errno));
		nRet = eLIB_ERR_DRIVER;
	}

	return nRet;
}

/** 
 * @brief データリード関数
 * 
 * @param[in]  fd			file descriptor.
 * @param[out] *pReadData	リードデータ格納先
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DIRECTION		direction error.
 * @retval eLIB_ERR_DRIVER			driver call error.
 * @retval eLIB_ERR_PARAM			parameter error.
 *
 * MPIOLVLからデータをリードする。@n
 * 方向設定がすべて出力の場合は、エラーとなる。
 *
 */
int ReadPort(int fd, unsigned char *pReadData)
{
	int nRet;
	unsigned char ucDirection = 0;
	unsigned char ucDeviceID = 0;
	IOPORT_UCHAR	bIOPort;

	/* パラメーターチェック */
	if(NULL == pReadData)
	{
		return eLIB_ERR_PARAM;
	}

	/* Directionのチェック */
	nRet = CheckDirection(fd, &ucDirection);
	if(0 > nRet)
	{
		return nRet;
	}

	if(0x00 == (ucDirection & 0x3F))
	{
		return eLIB_ERR_DIRECTION;
	}

	/* MPIOINV へ0x00を書き込むことでビット反転させない */
	nRet = SetIntEdge(fd, 0x00);
	if(0 > nRet)
	{
		return nRet;
	}

	/* DVID からDeviceIDを取得する */
	bIOPort.Offset = DVID;
	bIOPort.Value = 0x00;	/* Dummy */
	nRet = ioctl(fd, IOCTL_DIO_READ_PORT, &bIOPort);
	if(0 > nRet)
	{
		fprintf(stderr, "ioctl IOCTL_DIO_READ_PORT error : <%s>\n", strerror(errno));
		return eLIB_ERR_DRIVER;
	}
	ucDeviceID = bIOPort.Value;

	/* MPIOLVL からデータを取得する */
	bIOPort.Offset = MPIOLVL;
	bIOPort.Value = 0x00;	/* Dummy */
	nRet = ioctl(fd, IOCTL_DIO_READ_PORT, &bIOPort);
	if(0 > nRet)
	{
		fprintf(stderr, "ioctl IOCTL_DIO_READ_PORT error : <%s>\n", strerror(errno));
		return eLIB_ERR_DRIVER;
	}

	bIOPort.Value = (bIOPort.Value & 0x3F) | (GetBoardID(ucDeviceID, bIOPort.Value) << 6);

	*pReadData = bIOPort.Value;

	DBG_PRINTF("[%s(%d)] OU nRet = %d Read = 0x%02x\n", __func__, __LINE__, nRet, bIOPort.Value);

	return eLIB_OK;
}

/** 
 * @brief データライト関数
 * 
 * @param[in] fd				file descriptor.
 * @param[in] WriteData			出力データ
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DIRECTION		direction error.
 * @retval eLIB_ERR_DRIVER			driver call error.
 *
 * MPIOLVLにデータを出力する。@n
 * 方向設定がすべて入力の場合は、エラーとなる。
 *
 */
int WritePort(int fd, unsigned char WriteData)
{
	int nRet;
	unsigned char ucDirection = 0;
	IOPORT_UCHAR bIOPort;

	/* Directionのチェック */
	nRet = CheckDirection(fd, &ucDirection);
	if(0 > nRet)
	{
		return nRet;
	}

	if(0x3F == (ucDirection & 0x3F))
	{
		return eLIB_ERR_DIRECTION;
	}

	/* MPIOINV へ0x00を書き込むことでビット反転させない */
	nRet = SetIntEdge(fd, 0x00);
	if(0 > nRet)
	{
		return nRet;
	}

	/* MPIOLVL へデータを出力する */
	bIOPort.Offset = MPIOLVL;
	bIOPort.Value = WriteData;
	nRet = ioctl(fd, IOCTL_DIO_WRITE_PORT, &bIOPort);
	if(0 > nRet)
	{
		fprintf(stderr, "ioctl IOCTL_DIO_WRITE_PORT error : <%s>\n", strerror(errno));
		return eLIB_ERR_DRIVER;
	}

	DBG_PRINTF("[%s(%d)] OU nRet = %d Write = 0x%02x\n", __func__, __LINE__, nRet, bIOPort.Value);

	return eLIB_OK;
}

/** 
 * @brief 割り込み待ち開始
 * 
 * @param[in] fd			file descriptor.
 * @param[in] IntBit		割り込みを有効にするBit
 * @param[in] IntEdge		割り込みを検出するレベル設定(0:立ち下がり 1:立ち上り)
 * @param[in] pCallBack		割り込み検出時に呼び出すコールバック関数を指定
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DIRECTION		direction error.
 * @retval eLIB_ERR_DRIVER			driver call error.
 * @retval eLIB_ERR_PARAM			parameter error.
 * @retval eLIB_ERR_INTERRUPT		already interrupt waiting.
 * @retval eLIB_ERR_TIMEOUT			interrupt timeout/cancel (Blocking)
 *
 * デバイスごとに1つの割り込み待ちが可能。すでに割り込み待ち実行中の場合はエラー終了する。@n
 * 指定したBitで割り込みを検出した場合、コールバック関数を呼び出す。@n
 * 割り込みの検出が終了するまで、file descriptorをクローズした場合は割り込み待ちを終了する。@n
 * pCallBackにNULLを設定した場合はブロッキング動作を行い、割り込みを受信/タイムアウト/InterruptStopのいずれかにより処理を終了する。
 *
 */
int InterruptStart(int fd, unsigned char IntBit, unsigned char IntEdge, PAPIFUNC pCallBack)
{
	int nRet;
	int nRetValue = eLIB_OK;
	unsigned char ucDirection;
	IOPORT_UCHAR bIOPort;
	pthread_t thread_id;
	pinterrupt_ctrl_t pinterrupt_arg;

	/* パラメーターチェック */
	if(0 == IntBit)
	{
		return eLIB_ERR_PARAM;
	}

	sem_wait(&g_sem_ref_interrupt_arg);

	pinterrupt_arg = interrupt_args_get(fd);
	if(NULL == pinterrupt_arg)
	{
		nRetValue = eLIB_ERR_INTERRUPT;
		goto error_exit;
	}

	/* 既に割り込み待ち中 */
	if(1 == pinterrupt_arg->int_start)
	{
		nRetValue = eLIB_ERR_INTERRUPT;
		goto error_exit;
	}

	/* Directionのチェック */
	nRet = CheckDirection(fd, &ucDirection);
	if(0 > nRet)
	{
		nRetValue = nRet;
		goto error_exit;
	}

	if(!(IntBit & ucDirection))
	{
		nRetValue = eLIB_ERR_DIRECTION;
		goto error_exit;
	}

	/* 検出方向を設定 */
	nRet = SetIntEdge(fd, ~IntEdge);
	if(0 > nRet)
	{
		nRetValue = nRet;
		goto error_exit;
	}

	/* MPIOINT 割り込みを有効化 */
	bIOPort.Offset = MPIOINT;
	bIOPort.Value = IntBit;
	nRet = ioctl(fd, IOCTL_DIO_WRITE_PORT, &bIOPort);
	if(0 > nRet)
	{
		fprintf(stderr, "ioctl IOCTL_DIO_WRITE_PORT error : <%s>\n", strerror(errno));
		nRetValue = eLIB_ERR_DRIVER;
		goto error_exit;
	}

	pinterrupt_arg->pFunc = pCallBack;
	pinterrupt_arg->int_start = 1;
	sem_trywait(&pinterrupt_arg->semIntComplete);

	if(NULL == pCallBack)
	{
		sem_post(&g_sem_ref_interrupt_arg);
		nRet = ioctl(fd, IOCTL_DIO_START_INTERRUPT, NULL);
		if(0 > nRet)
		{
			fprintf(stderr, "ioctl IOCTL_DIO_START_INTERRUPT error : <%s>\n", strerror(errno));
			nRetValue = eLIB_ERR_TIMEOUT;
		}

		/* 割り込み待ちフラグ完了 */
		sem_wait(&g_sem_ref_interrupt_arg);
		pinterrupt_arg->int_start = 0;
		sem_post(&pinterrupt_arg->semIntComplete);
	}
	else
	{
		/* 割り込み待ちスレッド開始 */
		nRet = pthread_create(&thread_id, NULL, InterruptThread, (void *)pinterrupt_arg);
		if(0 > nRet)
		{
			DBG_FPRINTF(stderr, "[%s(%d)] pthread_create ERROR : %s\n", __func__, __LINE__, strerror(nRet));
			nRetValue = eLIB_ERR_INTERRUPT;
		}
		else
		{
			DBG_PRINTF("[%s(%d)] thread_id : %lu\n", __func__, __LINE__, thread_id);
			pinterrupt_arg->thread_id = thread_id;
		}
	}

error_exit:
	sem_post(&g_sem_ref_interrupt_arg);

	return nRetValue;
}

/** 
 * @brief 割り込み待ち終了処理
 * 
 * @param[in] fd		file descriptor.
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DRIVER			driver call error.
 * @retval eLIB_ERR_INTERRUPT		already interrupt waiting.
 * @retval eLIB_ERR_TIMEOUT			timeout.
 *
 * 割り込み待ちを終了して、コールバックをコールせずに終了する。@n
 * 終了しない場合は、割り込み終了待ち1分でタイムアウトする。
 *
 */
int InterruptStop(int fd)
{
	int nRet;
	pinterrupt_ctrl_t pinterrupt_arg;
	struct timespec timeout;

	sem_wait(&g_sem_ref_interrupt_arg);
	pinterrupt_arg = interrupt_args_get(fd);
	if(NULL == pinterrupt_arg)
	{
		sem_post(&g_sem_ref_interrupt_arg);
		return eLIB_ERR_INTERRUPT;
	}

	/* 割り込み待ち動作中の場合は終了待ちを行う */
	if(1 == pinterrupt_arg->int_start)
	{
		pinterrupt_arg->pFunc = NULL;
		sem_post(&g_sem_ref_interrupt_arg);

		nRet = ioctl(fd, IOCTL_DIO_STOP_INTERRUPT, NULL);
		if(0 > nRet)
		{
			fprintf(stderr, "ioctl IOCTL_DIO_STOP_INTERRUPT error : <%s>\n", strerror(errno));
			return eLIB_ERR_DRIVER;
		}

		clock_gettime(CLOCK_REALTIME, &timeout);
		timeout.tv_sec += 60;
		if(sem_timedwait(&pinterrupt_arg->semIntComplete, &timeout))
		{
			if(ETIMEDOUT == errno)
			{
				return eLIB_ERR_TIMEOUT;
			}
			else
			{
				fprintf(stderr, "sem_timedwait error : <%s>\n", strerror(errno));
				abort();
			}
		}
	}
	else
	{
		sem_post(&g_sem_ref_interrupt_arg);
	}

	return eLIB_OK;
}

/** 
 * @brief 割り込み待ち開始（SIGNAL）
 * 
 * @param[in] fd			file descriptor.
 * @param[in] IntBit		割り込みを有効にするBit
 * @param[in] IntEdge		割り込みを検出するレベル設定(0:立ち下がり 1:立ち上り)
 * @param[in] nSignal		通知するシグナル番号
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DRIVER			driver call error.
 * @retval eLIB_ERR_PARAM			parameter error.
 *
 * デバイスごとに1つの割り込み待ち可能。すでに割り込み待ち実行中の場合はエラー終了する。@n
 * 指定したBitで割り込みを検出した場合、指定したシグナル番号に通知を行う。@n
 * 割り込みの検出が終了するまで、file descriptorをクローズした場合は割り込み待ちを終了する。
 *
 */
int InterruptStartSignal(int fd, unsigned char IntBit, unsigned char IntEdge, int nSignal)
{
	T_IOCTL_DIO_INT_ENABLE	sEnable;
	int nRet;
	unsigned char ucDirection;

	/* パラメーターチェック */
	if(0 == IntBit)
	{
		return eLIB_ERR_PARAM;
	}

	/* Directionのチェック */
	nRet = CheckDirection(fd, &ucDirection);
	if(0 > nRet)
	{
		return nRet;
	}

	if(!(IntBit & ucDirection))
	{
		return eLIB_ERR_DIRECTION;
	}

	/* 検出方向を設定 */
	nRet = SetIntEdge(fd, ~IntEdge);
	if(0 > nRet)
	{
		return nRet;
	}

	sEnable.ucIntBit = IntBit;
	sEnable.nSignalNo = nSignal;

	nRet = ioctl(fd, REX_DIO_IOCS_INTSIG_ENABLE, &sEnable);
	if(0 > nRet)
	{
		fprintf(stderr, "ioctl REX_DIO_IOCS_INTSIG_ENABLE error : <%s>\n", strerror(errno));
		return eLIB_ERR_DRIVER;
	}

	return eLIB_OK;
}

/** 
 * @brief 割り込み待ち終了（SIGNAL）
 * 
 * @param[in] fd			file descriptor.
 * @param[in] nSignal		通知するシグナル番号
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DRIVER			driver call error.
 *
 * ドライバーの割り込み待ちを解除する。@n
 * 既に割り込み発生によりシグナルが通知された場合はドライバーは割込み待ち解除されている。
 *
 */
int InterruptStopSignal(int fd, int nSignal)
{
	T_IOCTL_DIO_INT_DISABLE	sDisable;
	int nRet;

	sDisable.nSignalNo = nSignal;
	nRet = ioctl(fd, REX_DIO_IOCS_INTSIG_DISABLE, &sDisable);
	if(0 > nRet)
	{
		fprintf(stderr, "ioctl REX_DIO_IOCS_INTSIG_DISABLE error : <%s>\n", strerror(errno));
		return eLIB_ERR_DRIVER;
	}

	return eLIB_OK;
}

/* ------------------------ */
/*	PRIVATE FUNCTION		*/
/* ------------------------ */
/** 
 * @brief 方向設定の取得関数
 * 
 * @param[in]  fd			file descriptor.
 * @param[out] *pDirection	取得値の格納先
 * 
 * @retval 0		success.
 * @retval -2		driver call error.
 * @retval -3		parameter error.
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DRIVER			driver call error.
 * @retval eLIB_ERR_PARAM			parameter error.
 *
 * MPIOSELから現在の方向設定を取得する。
 *
 */
static int CheckDirection(int fd, unsigned char *pDirection)
{
	int nRet;
	IOPORT_UCHAR	bIOPort;

	/* パラメーターチェック */
	if(NULL == pDirection)
	{
		nRet = eLIB_ERR_PARAM;
	}
	else
	{
		bIOPort.Offset = MPIOSEL;
		bIOPort.Value = 0x00;		/* Dummy */
		nRet = ioctl(fd, IOCTL_DIO_READ_PORT, &bIOPort);
		if(0 > nRet)
		{
			fprintf(stderr, "ioctl IOCTL_DIO_READ_PORT error : <%s>\n", strerror(errno));
			nRet = eLIB_ERR_DRIVER;
		}
		else
		{
			*pDirection = bIOPort.Value;
		}
	}

	return nRet;
}

/** 
 * @brief データ反転設定関数
 * 
 * @param[in] fd			file descriptor.
 * @param[in] IntEdge		レベル設定
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DRIVER			driver call error.
 *
 * MPIOINVにデータを設定することで、対応するBitのデータの反転設定を行う。@n
 * 0を設定することで、生データとなる。@n
 * 通常は立上りを立下がりの割り込み検出に変更する場合に使用する。
 *
 */
static int SetIntEdge(int fd, unsigned char IntEdge)
{
	int nRet;
	IOPORT_UCHAR	bIOPort;

	bIOPort.Offset = MPIOINV;
	bIOPort.Value = IntEdge;
	nRet = ioctl(fd, IOCTL_DIO_WRITE_PORT, &bIOPort);
	if(0 > nRet)
	{
		fprintf(stderr, "ioctl IOCTL_DIO_WRITE_PORT error : <%s>\n", strerror(errno));
		return eLIB_ERR_DRIVER;
	}

	return eLIB_OK;
}

/** 
 * @brief ボードIDを算出する
 * 
 * @param [in]	dvid	DVID
 * @param [in]	value	MPIOLVL
 * 
 * @return ボードID
 *
 * MPIOLVLの値からボードIDを算出する。@n
 * MPIOLVLの値はMPIOINVの値が0x00の状態のものを指定する。
 */
static unsigned char GetBoardID(unsigned char dvid, unsigned char value)
{
	unsigned char dev = 0;
	unsigned char id[2][4] = {
		{ 0x03, 0x01, 0x02, 0x00},
		{ 0x03, 0x02, 0x01, 0x00}
	};

	if(	(0x24 == dvid)		/*	c154	*/
	||	(0x84 == dvid))		/*	v354	*/
	{
		dev = 1;
	}

	return id[dev][((value & 0xc0) >> 6)];
}

/** 
 * @brief 割り込み待ちスレッド
 * 
 * @param[in] arg	interrupt_ctrl_tのポインタ
 * 
 * @return void
 */
static void *InterruptThread(void *arg)
{
	int nRet;
	int fd;
	pinterrupt_ctrl_t pinterrupt_arg = (pinterrupt_ctrl_t)arg;

	DBG_PRINTF("[%s(%d)] IN : \n", __func__, __LINE__);

	/* detach */
	nRet = pthread_detach(pthread_self());
	if(0 != nRet)
	{
		DBG_FPRINTF(stderr, "[%s(%d)] thread detach ERROR : %s\n", __func__, __LINE__, strerror(nRet));
	}
	else
	{
		sem_wait(&g_sem_ref_interrupt_arg);
		fd = pinterrupt_arg->fd;
		sem_post(&g_sem_ref_interrupt_arg);

		/* 割り込み待ち開始 */
		nRet = ioctl(fd, IOCTL_DIO_START_INTERRUPT, NULL);
		if(0 > nRet)
		{
			fprintf(stderr, "ioctl IOCTL_DIO_START_INTERRUPT error : <%s>\n", strerror(errno));
		}
		else
		{
			sem_wait(&g_sem_ref_interrupt_arg);
			/* コールバック関数の呼び出し */
			if(NULL != pinterrupt_arg->pFunc)
			{
				(pinterrupt_arg->pFunc)();
			}
			sem_post(&g_sem_ref_interrupt_arg);
		}
	}

	/* 割り込み待ちフラグ完了 */
	sem_wait(&g_sem_ref_interrupt_arg);
	pinterrupt_arg->int_start = 0;
	sem_post(&g_sem_ref_interrupt_arg);
	sem_post(&pinterrupt_arg->semIntComplete);

	DBG_PRINTF("[%s(%d)] OU : \n", __func__, __LINE__);

	return NULL;
}

/** 
 * @brief 割り込みスレッド管理変数初期化
 * 
 * @param [in]	fd		file descriptor.
 * @param [in]	id			board id.
 * @param [in]	kind		board kind.
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_INTERRUPT		already used.
 *
 * 割り込み動作用スレッド管理用の変数を初期化する。@n
 * １デバイスにつき１つだけ使用可能。
 */
static int interrupt_args_add(int fd, int id, int kind)
{
	int nRet = eLIB_OK;
	int index = id * kind;
	pinterrupt_ctrl_t pinterrupt_arg = &g_interrupt_args[index];

	sem_wait(&g_sem_ref_interrupt_arg);
	if(-1 == pinterrupt_arg->fd)
	{
		pinterrupt_arg->fd = fd;
		pinterrupt_arg->pFunc = NULL;
		pinterrupt_arg->int_start = 0;
		sem_init(&pinterrupt_arg->semIntComplete, 0, 1);
	}
	else
	{
		nRet = eLIB_ERR_INTERRUPT;
	}
	sem_post(&g_sem_ref_interrupt_arg);

	return nRet;
}

/** 
 * @brief 割り込みスレッド管理変数初期化
 * 
 * @param [in]	fd		file descriptor.
 * 
 * @retval eLIB_OK					success.
 * @retval eLIB_ERR_DRIVER			driver call error.
 *
 * DioCloseする際に使用していた領域を初期化する。
 */
static int interrupt_args_clear(int fd)
{
	int nRet = eLIB_OK;
	pinterrupt_ctrl_t pinterrupt_arg;

	sem_wait(&g_sem_ref_interrupt_arg);
	pinterrupt_arg = interrupt_args_get(fd);
	if(NULL == pinterrupt_arg)
	{
		nRet = eLIB_ERR_DRIVER;
	}
	else
	{
		/*	変数初期化	*/
		pinterrupt_arg->fd = -1;
		pinterrupt_arg->pFunc = NULL;
		pinterrupt_arg->int_start = 0;
		sem_destroy(&pinterrupt_arg->semIntComplete);
	}

	sem_post(&g_sem_ref_interrupt_arg);

	return nRet;
}

/** 
 * @brief file descriptorに該当するスレッド管理変数を取得
 * 
 * @param [in]	fd		file descriptor.
 * 
 * @return 管理変数のポインタ
 *
 * DioOpen時に対応付けたfile descriptorに該当するものがある場合は@n
 * その領域のアドレスを返す。@n
 * 関数内ではセマフォは使用しないため呼び出し元で制御する。
 */
static pinterrupt_ctrl_t interrupt_args_get(int fd)
{
	int i;
	pinterrupt_ctrl_t pinterrupt_arg = NULL;

	for(i = 0; i < MAX_DEVICE_NUM; i++)
	{
		if(fd == g_interrupt_args[i].fd)
		{
			pinterrupt_arg = &g_interrupt_args[i];
			break;
		}
	}

	return pinterrupt_arg;
}

