/** 
 * @file dioapli.c
 * @brief DIO入出力サンプルアプリケーション
 * @version 1.0.1
 * @date 2012-11-14
 * @par Copyright
 *  (C) 2012 RATOC Systems, Inc.
 * @par History
 *  - 2012.02.17 v1.0.0
 *    -# Initial Version.
 *
 *  - 2012.11.14 v1.0.1
 *    -# for glib 2.32
 */
/*
 *  Copyright (C) 2012 RATOC Systems, Inc.
 *  Dual licensed under the BSD (BSD-LICENSE.txt)
 *  and GPL (GPL-LICENSE.txt) licenses.
 */

#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <gtk/gtk.h>

#include <rexpx6x70d_dio.h>

/* ------------------------ */
/* DEFINE					*/
/* ------------------------ */
/** 
 * @enum	interrupt_mode
 * @brief	割込み待ちモード制御の管理に使用する
 */
typedef enum interrupt_mode {
	eINT_MODE_STOP		= -1,								/**< 割込み停止中									*/
	eINT_MODE_START		= -2,								/**< 割込み動作中									*/
	eINT_MODE_RECEIVE	= -3,								/**< 割込み受信										*/
	eINT_MODE_CANCEL	= -4,								/**< 割込みキャンセル								*/
} interrupt_mode_e;											/**< 割込み動作モード								*/

/** 
 * @var		g_int_flg
 * @brief	割込みのシグナルハンドラからの通知用
 */
volatile sig_atomic_t g_int_flg = eINT_MODE_STOP;			/**< 割り込み通知用									*/

#define SIG_DIO_INTERRUPT		(SIGRTMIN + 1)				/**< 割り込み用シグナル番号							*/

/** 
 * @name 入出力方向
 * @{ */
#define DIR_OUTPUT		(0)									/**< 出力方向										*/
#define DIR_INPUT		(1)									/**< 入力方向										*/
/**  @} */

/** 
 * @name 割り込み検出レベル
 * @{ */
#define INT_LEVEL_LOW	(0)									/**< 割り込み検出レベル：Low						*/
#define INT_LEVEL_HIGH	(1)									/**< 割り込み検出レベル：High						*/
/**  @} */

/** 
 * @name 割り込みON/OFF
 * @{ */
#define INT_OFF			(0)									/**< 割り込みOFF									*/
#define INT_ON			(1)									/**< 割り込みON										*/
/**  @} */

/** 
 * @name 割り込み検出レベルコンボボックス設定値順
 * @{ */
#define COMBO_INT_NONE	(0)									/**< 設定無し										*/
#define COMBO_INT_HIGH	(1)									/**< HIGH											*/
#define COMBO_INT_LOW	(2)									/**< LOW											*/
/**  @} */

#define MAX_BIT_NUM		(6)									/**< DIO入出力ビット数								*/

/* ------------------------ */
/*	STRUCT					*/
/* ------------------------ */
/** 
 * @struct struct DIO
 * @brief DIO設定値
 */
typedef struct DIO {
	unsigned char kind;										/**< ボード種別										*/
	unsigned char id;										/**< ボードID										*/
	unsigned char direction;								/**< 入出力方向										*/
	unsigned char interrupt_bit;							/**< 割り込みON/OFF									*/
	unsigned char interrupt_level;							/**< 割り込み検出レベル								*/
} tagDIO;

typedef void (*FPFunc)(GtkButton *, gpointer);

/* ------------------------ */
/*	prottype				*/
/* ------------------------ */
static void radiobtn_onclicked_bit5(GtkWidget *widget, gpointer data);
static void radiobtn_onclicked_bit4(GtkWidget *widget, gpointer data);
static void radiobtn_onclicked_bit3(GtkWidget *widget, gpointer data);
static void radiobtn_onclicked_bit2(GtkWidget *widget, gpointer data);
static void radiobtn_onclicked_bit1(GtkWidget *widget, gpointer data);
static void radiobtn_onclicked_bit0(GtkWidget *widget, gpointer data);
static void all_set_btn_onclicked(GtkWidget *widget, gpointer data);

static void destroy(GtkWidget *widget, gpointer data);
static void bord_kind_combo_box_onchanged(GtkComboBox *widget, gpointer data);
static void bord_id_combo_box_onchanged(GtkComboBox *widget, gpointer data);
static void interrupt_level_combo_box_onchanged(GtkComboBox *widget, gpointer data);
static void write_btn_onclicked(GtkWidget *widget, gpointer data);
static void read_btn_onclicked(GtkWidget *widget, gpointer data);
static void interrupt_btn_onclicked(GtkWidget *widget, gpointer data);
static int driver_close_force(void);
static int driver_close(int fd);
static int driver_open(int id, int kind);
static void create_window(void);
static void *interrupt_waiting_thread(void *arg);
static void interrupt_signal_handler(int signo);
static void interrupt_thread_stop(void);

/* ------------------------ */
/*	GLOBAL VARIABLE			*/
/* ------------------------ */
/** 
 * @var rc_func
 * @brief ラジオボタンクリック時のイベント関数
 *
 * bit5 - bit0に対応
 */
FPFunc rc_func[MAX_BIT_NUM] = {
	(FPFunc)radiobtn_onclicked_bit5,
	(FPFunc)radiobtn_onclicked_bit4,
	(FPFunc)radiobtn_onclicked_bit3,
	(FPFunc)radiobtn_onclicked_bit2,
	(FPFunc)radiobtn_onclicked_bit1,
	(FPFunc)radiobtn_onclicked_bit0,
};

/** 
 * @var g_fd
 * @brief ファイルディスクリプタ
 */
int g_fd = -1;

/** 
 * @var g_driver_open_cnt
 * @brief ドライバ多重Openカウンタ
 */
volatile int g_driver_open_cnt = 0;

/** 
 * @var stDio
 * @brief DIO設定値
 */
tagDIO stDio = {
	.kind            = 0x00,
	.id              = 0x00,
	.direction       = 0x3f,
	.interrupt_bit   = 0x00,
	.interrupt_level = 0x00,
};

/** 
 * @name 設定値set/get必要なウィジット
 * @{ */
GtkWidget *gDirDataIoRadioBtn[MAX_BIT_NUM][2];

GtkWidget *g_button_int;
GtkWidget *g_label_info;
GtkWidget *g_label_read_data;
GtkWidget *g_label_int_data;

GtkWidget *g_entry_write_data;
/**  @} */

/* ------------------------ */
/*	FUNCTION				*/
/* ------------------------ */
/** 
 * @brief ウィンドウアプリケーション終了
 * 
 * @param[in] widget
 * @param[in] data
 *
 * @return		なし
 */
static void destroy(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

/** 
 * @brief ボード種別コンボボックス変更時の処理
 * 
 * @param[in] widget
 * @param[in] data
 *
 * @return		なし
 *
 * ボード種別を設定する
 *
 */
static void bord_kind_combo_box_onchanged(GtkComboBox *widget, gpointer data)
{
#ifdef DEBUG
	gchar buf[16];
#endif

	driver_close_force();
	stDio.kind = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
	gtk_label_set_text(GTK_LABEL (g_label_info), "");

#ifdef DEBUG
	sprintf(buf, "%d", stDio.kind);
	gtk_label_set_text(GTK_LABEL (g_label_info), buf);
#endif
}

/** 
 * @brief ボードIDコンボボックス変更時の処理
 * 
 * @param[in] widget
 * @param[in] data
 *
 * @return		なし
 *
 * ボードIDを設定する
 *
 */
static void bord_id_combo_box_onchanged(GtkComboBox *widget, gpointer data)
{
#ifdef DEBUG
	gchar buf[16];
#endif

	driver_close_force();
	stDio.id = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
	gtk_label_set_text(GTK_LABEL (g_label_info), "");

#ifdef DEBUG
	sprintf(buf, "%d", stDio.id);
	gtk_label_set_text(GTK_LABEL (g_label_info), buf);
#endif
}

/** 
 * @brief 割り込みレベルコンボボックス変更時の処理
 * 
 * @param[in] widget	combo box.
 * @param[in] data		変更されたBit(0 - 5)
 *
 * @return		なし
 *
 * 対象のBitの割り込みON/OFFと割り込みレベルの設定
 *
 */
static void interrupt_level_combo_box_onchanged(GtkComboBox *widget, gpointer data)
{
	int bit = (int)data;
	gint index;
#ifdef DEBUG
	gchar buf[32];
#endif

	index = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
	switch (index)
	{
	case COMBO_INT_NONE:		/* 割り込みOFF		*/
		stDio.interrupt_bit &= ~(1 << bit);
		stDio.interrupt_level &= ~(1 << bit);
		break;
	case COMBO_INT_LOW:			/* 割り込みON、割り込みレベルLow	*/
		stDio.interrupt_bit |= (1 << bit);
		stDio.interrupt_level &= ~(1 << bit);
		break;
	case COMBO_INT_HIGH:		/* 割り込みON、割り込みレベルHigh	*/
		stDio.interrupt_bit |= (1 << bit);
		stDio.interrupt_level |= (1 << bit);
		break;
	default:
		break;
	}

#ifdef DEBUG
	sprintf(buf, "bit : 0x%02x, level : 0x%02x", stDio.interrupt_bit, stDio.interrupt_level);
	gtk_label_set_text(GTK_LABEL (g_label_info), buf);
#endif
}

/** 
 * @brief DIO入出力方向の設定
 * 
 * @param[in] bit	変更するBit(0-5)
 * @param[in] dir	方向(DIR_INPUT/DIR_OUTPUT)
 *
 * @return		なし
 */
static void set_dir_io_data(char bit, char dir)
{
#ifdef DEBUG
	gchar buf[16];
#endif

	if(DIR_INPUT == dir)
	{
		stDio.direction |= (1 << bit);
	}
	else
	if(DIR_OUTPUT == dir)
	{
		stDio.direction &= ~(1 << bit);
	}
	else
	{
		gtk_label_set_text(GTK_LABEL (g_label_info), "Internal Error");
	}

#ifdef DEBUG
	sprintf(buf, "%d%d, 0x%02x", bit, dir, stDio.direction);
	gtk_label_set_text(GTK_LABEL (g_label_info), buf);
#endif
}

/** 
 * @brief Bit5の入出力ラジオボタン変更時の処理
 * 
 * @param[in] widget	クリックされたラジオボタン
 * @param[in] data		方向(DIR_INPUT/DIR_OUTPUT)
 *
 * @return		なし
 *
 */
static void radiobtn_onclicked_bit5(GtkWidget *widget, gpointer data)
{
	switch((int)data)
	{
	case DIR_INPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(5, DIR_INPUT);
		}
		break;
	case DIR_OUTPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(5, DIR_OUTPUT);
		}
		break;
	default:
		break;
	}
}

/** 
 * @brief Bit4の入出力ラジオボタン変更時の処理
 * 
 * @param[in] widget	クリックされたラジオボタン
 * @param[in] data		方向(DIR_INPUT/DIR_OUTPUT)
 *
 * @return		なし
 *
 */
static void radiobtn_onclicked_bit4(GtkWidget *widget, gpointer data)
{
	switch((int)data)
	{
	case DIR_INPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(4, DIR_INPUT);
		}
		break;
	case DIR_OUTPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(4, DIR_OUTPUT);
		}
		break;
	default:
		break;
	}
}

/** 
 * @brief Bit3の入出力ラジオボタン変更時の処理
 * 
 * @param[in] widget	クリックされたラジオボタン
 * @param[in] data		方向(DIR_INPUT/DIR_OUTPUT)
 *
 * @return		なし
 *
 */
static void radiobtn_onclicked_bit3(GtkWidget *widget, gpointer data)
{
	switch((int)data)
	{
	case DIR_INPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(3, DIR_INPUT);
		}
		break;
	case DIR_OUTPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(3, DIR_OUTPUT);
		}
		break;
	default:
		break;
	}
}

/** 
 * @brief Bit2の入出力ラジオボタン変更時の処理
 * 
 * @param[in] widget	クリックされたラジオボタン
 * @param[in] data		方向(DIR_INPUT/DIR_OUTPUT)
 *
 * @return		なし
 *
 */
static void radiobtn_onclicked_bit2(GtkWidget *widget, gpointer data)
{
	switch((int)data)
	{
	case DIR_INPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(2, DIR_INPUT);
		}
		break;
	case DIR_OUTPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(2, DIR_OUTPUT);
		}
		break;
	default:
		break;
	}
}

/** 
 * @brief Bit1の入出力ラジオボタン変更時の処理
 * 
 * @param[in] widget	クリックされたラジオボタン
 * @param[in] data		方向(DIR_INPUT/DIR_OUTPUT)
 *
 * @return		なし
 *
 */
static void radiobtn_onclicked_bit1(GtkWidget *widget, gpointer data)
{
	switch((int)data)
	{
	case DIR_INPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(1, DIR_INPUT);
		}
		break;
	case DIR_OUTPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(1, DIR_OUTPUT);
		}
		break;
	default:
		break;
	}
}

/** 
 * @brief Bit0の入出力ラジオボタン変更時の処理
 * 
 * @param[in] widget	クリックされたラジオボタン
 * @param[in] data		方向(DIR_INPUT/DIR_OUTPUT)
 *
 * @return		なし
 *
 */
static void radiobtn_onclicked_bit0(GtkWidget *widget, gpointer data)
{
	switch((int)data)
	{
	case DIR_INPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(0, DIR_INPUT);
		}
		break;
	case DIR_OUTPUT:
    	if(TRUE == gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
		{
			set_dir_io_data(0, DIR_OUTPUT);
		}
		break;
	default:
		break;
	}
}

/** 
 * @brief 全入力/全出力設定ボタンクリック時の処理
 * 
 * @param[in] widget	button
 * @param[in] data		押下されたボタン(DIR_INPUT/DIR_OUTPUT)
 *
 * @return		なし
 *
 * 押下された側のラジオボタンをアクティブに一括設定する
 *
 */
static void all_set_btn_onclicked(GtkWidget *widget, gpointer data)
{
	int i;
	int dir = (int)data;

	for(i = 0; i < MAX_BIT_NUM; i++)
	{
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (gDirDataIoRadioBtn[i][dir]), TRUE);
	}
}

/** 
 * @brief ライトボタンクリック時の処理
 * 
 * @param[in] widget	button
 * @param[in] data		NULL
 *
 * @return		なし
 *
 * テキスト領域に入力された値をDIO出力する
 *
 */
static void write_btn_onclicked(GtkWidget *widget, gpointer data)
{
	const gchar *buftmp;
	gchar cMsg[64];
	long lHex = 0;
	int fd;
	int nRet;

	/* 入力値を16進数として数値に変換 */
	buftmp = gtk_entry_get_text (GTK_ENTRY (g_entry_write_data));
	lHex = strtol(buftmp, NULL, 16);

	/* DioOpen */
	fd = driver_open(stDio.id, stDio.kind);
	if(0 > fd)
	{
		return;
	}

	/* SetDirection */
	nRet = SetDirection(fd, stDio.direction);
	if(0 > nRet)
	{
		sprintf(cMsg, "SetDirection Error : %d", nRet);
		gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
		/* DioClose */
		driver_close(fd);
		return;
	}

	/* WritePort */
	nRet = WritePort(fd, lHex);
	if(0 > nRet)
	{
		sprintf(cMsg, "WritePort Error : %d", nRet);
		gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
		/* DioClose */
		driver_close(fd);
		return;
	}

	memset(cMsg, 0x00, sizeof(cMsg));
	gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);

	/* DioClose */
	driver_close(fd);
}

/** 
 * @brief リードボタンクリック時の処理
 * 
 * @param[in] widget	button
 * @param[in] data		NULL
 *
 * @return		なし
 *
 * DIOから値を取得して出力する
 *
 */
static void read_btn_onclicked(GtkWidget *widget, gpointer data)
{
	unsigned char ucReadData = 0;
	gchar cMsg[64];
	int fd;
	int nRet;

	memset(cMsg, 0x00, sizeof(cMsg));
	gtk_label_set_text (GTK_LABEL (g_label_read_data), cMsg);

	/* DioOpen */
	fd = driver_open(stDio.id, stDio.kind);
	if(0 > fd)
	{
		return;
	}

	/* SetDirection */
	nRet = SetDirection(fd, stDio.direction);
	if(0 > nRet)
	{
		sprintf(cMsg, "SetDirection Error : %d", nRet);
		gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
		/* DioClose */
		driver_close(fd);
		return;
	}

	/* ReadPort */
	nRet = ReadPort(fd, &ucReadData);
	if(0 > nRet)
	{
		sprintf(cMsg, "ReadPort Error : %d", nRet);
		gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
		/* DioClose */
		driver_close(fd);
		return;
	}

	/* DioClose */
	driver_close(fd);

	sprintf(cMsg, "0x%02x", ucReadData);
	gtk_label_set_text (GTK_LABEL (g_label_read_data), cMsg);
	gtk_label_set_text(GTK_LABEL (g_label_info), "");
}

/** 
 * @brief 割込み検出時のシグナルハンドラー
 * 
 * @param [in]	signo		signal no.
 *
 * 割込み検出した場合にドライバーからシグナル送信される。@n
 * 通知用のフラグを設定するのみ。
 */
static void interrupt_signal_handler(int signo)
{
	if(SIG_DIO_INTERRUPT == signo)
	{
		g_int_flg = eINT_MODE_RECEIVE;
	}
}

/** 
 * @brief 割込み検出待ちスレッド
 * 
 * @param [in]	arg			file descriptor.
 * 
 * @return NULL
 *
 * 割込み検出/中止時の処理を行う。@n
 *
 */
static void *interrupt_waiting_thread(void *arg)
{
	int ret;
	unsigned char ucReadData = 0;
	gchar cMsg[64];
	int nRet;
	int fd = (int)arg;

	ret = pthread_detach(pthread_self());
	if(0 != ret)
	{
		gdk_threads_enter();
		g_int_flg = eINT_MODE_STOP;
		driver_close(fd);
		gdk_flush();
		gdk_threads_leave();
	}
	else
	{
		while(eINT_MODE_START == g_int_flg)				/*	割込み検出	*/
		{
			usleep(100000);
		}

		gdk_threads_enter();
		if(eINT_MODE_RECEIVE != g_int_flg)				/*	割込み待ち中止時	*/
		{
			sprintf(cMsg, "割り込み中止");
			gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
		}
		else											/*	割込み受信	*/
		{
			/* ReadPort */
			nRet = ReadPort(fd, &ucReadData);
			if(0 > nRet)
			{
				sprintf(cMsg, "ReadPort Error : %d", nRet);
				gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
				/* DioClose */
				driver_close(fd);
			}
			else
			{
				sprintf(cMsg, "割り込み発生");
				gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);

				/* DioClose */
				driver_close(fd);

				sprintf(cMsg, "0x%02x", ucReadData);
				gtk_label_set_text (GTK_LABEL (g_label_int_data), cMsg);

				g_int_flg = eINT_MODE_STOP;
			}
		}
		gdk_flush();
		gdk_threads_leave();
	}

	return NULL;
}

/** 
 * @brief 割込み待ち停止処理
 *
 * 割込み動作中はドライバーの割込み処理の停止と割込み待ちスレッドの終了を行う。
 */
static void interrupt_thread_stop(void)
{
	if(eINT_MODE_STOP != g_int_flg)
	{
		gtk_button_set_label (GTK_BUTTON (g_button_int), "開始");
		if(eINT_MODE_START == g_int_flg)
		{
			InterruptStopSignal(g_fd, SIG_DIO_INTERRUPT);
		}
		g_int_flg = eINT_MODE_STOP;
	}
}

/** 
 * @brief 割り込みボタンクリック時の処理
 * 
 * @param[in] widget	button
 * @param[in] data		NULL
 *
 * @return		なし
 *
 * 割り込みON/OFFと割り込みレベルをドライバに設定する
 *
 */
static void interrupt_btn_onclicked(GtkWidget *widget, gpointer data)
{
	gchar cMsg[64];
	int fd;
	int nRet;
	pthread_t thread_id;

	gtk_label_set_text (GTK_LABEL (g_label_int_data), "");

	/* DioOpen */
	fd = driver_open(stDio.id, stDio.kind);
	if(0 > fd)
	{
		return;
	}

	if(	(0x00 == stDio.interrupt_bit)		/*	割込みBITすべて無効	*/
	&&	(eINT_MODE_STOP == g_int_flg))		/*	割込み停止中の場合	*/
	{
		gtk_label_set_text(GTK_LABEL (g_label_info), "Interrupt Condition Error");

		/* DioClose */
		driver_close(fd);
		return;
	}
	else
	if(eINT_MODE_STOP != g_int_flg)			/*	割込み動作中の場合	*/
	{
		/* 割り込み待ち終了 */
		driver_close_force();
		return;
	}

	/* SetDirection */
	nRet = SetDirection(fd, 0x3F);
	if(0 > nRet)
	{
		sprintf(cMsg, "SetDirection Error : %d", nRet);
		gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
		/* DioClose */
		driver_close(fd);
		return;
	}

	/* 入出力ラジオボタンも揃える */
	all_set_btn_onclicked(NULL, (gpointer *)DIR_INPUT);

	sprintf(cMsg, "割り込み待ち開始...");
	gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
	gtk_button_set_label ((GtkButton *)widget, "停止");

	g_int_flg = eINT_MODE_START;
	nRet = InterruptStartSignal(fd, stDio.interrupt_bit, stDio.interrupt_level, SIG_DIO_INTERRUPT);
	if(0 > nRet)
	{
		sprintf(cMsg, "InterruptStartSignal Error : %d", nRet);
		gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
		/* DioClose */
		driver_close(fd);
		return;
	}

	nRet = pthread_create(&thread_id, NULL, interrupt_waiting_thread, (void *)fd);
}

/** 
 * @brief ドライバオープン関数
 *
 * @param[in] boardid		ボードID
 * 
 * @return DioOpenの戻り値
 *
 * 複数回呼び出しても、実際にドライバオープンは初回のみ実行する。
 *
 */
static int driver_open(int id, int kind)
{
	gchar cMsg[64];

	if(-1 == g_fd)
	{
		/* DioOpen */
		g_fd = DioOpen(id, kind);
		if(0 > g_fd)
		{
			g_fd = -1;
			sprintf(cMsg, "Open Error : [ID : %d]", id);
			gtk_label_set_text(GTK_LABEL (g_label_info), cMsg);
		}
		else
		{
			g_driver_open_cnt ++;
		}
	}
	else
	{
		g_driver_open_cnt ++;
	}

	return g_fd;
}

/** 
 * @brief ドライバクローズ関数
 * 
 * @param[in] fd		file descriptor.
 * 
 * @return DioCloseの戻り値
 *
 * 実際にドライバクローズするには、driver_openを呼び出した回数分呼ぶ必要がある。
 *
 */
static int driver_close(int fd)
{
	int nRet;

	if(g_fd == fd)
	{
		g_driver_open_cnt --;
		if(0 == g_driver_open_cnt)
		{
			/* 割り込み待ち終了 */
			interrupt_thread_stop();

			/* DioClose */
			nRet = DioClose(fd);
			g_fd = -1;
		}
	}
	else
	{
		/* DioClose */
		nRet = DioClose(fd);
	}

	return nRet;
}

/** 
 * @brief ドライブクローズ関数（強制）
 * 
 * @return DioCloseの戻り値
 *
 * driver_open関数にて多重openしている場合でも、強制的にcloseする。
 *
 */
static int driver_close_force(void)
{
	int nRet = 0;

	/* 割り込み待ち終了 */
	interrupt_thread_stop();

	if(0 < g_driver_open_cnt)
	{
		nRet = DioClose(g_fd);
		g_driver_open_cnt = 0;
		g_fd = -1;
	}

	return nRet;
}

/** 
 * @brief ウィンドウ作成
 *
 * @return		なし
 */
static void create_window(void)
{
	GtkWidget *window;
	GtkWidget *label;
	GtkWidget *combo;
	GtkWidget *button;
	GtkWidget *vbox;
	GtkWidget *hbox;
	GtkWidget *frame;
	GtkWidget *lframe;
	GtkWidget *table;
	GtkWidget *inframe_vbox;

	gchar	cMsg[12] = {0};
	int		i = 0;
	
	/* メインウィンドウ */
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

	/* メインウィンドウ初期設定 */
	gtk_window_set_title (GTK_WINDOW (window), "REX-Px6xD/70D DIO Sample Application");
	gtk_container_set_border_width (GTK_CONTAINER (window), 5);

	g_signal_connect (G_OBJECT (window), "destroy",
                      G_CALLBACK (destroy), NULL);

#ifdef USE_GTK3
	vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 2);
#else
	vbox = gtk_vbox_new (FALSE, 2);
#endif
	gtk_container_add (GTK_CONTAINER (window), vbox);
	gtk_widget_show (vbox);

	/* ========================================================== */
#ifdef USE_GTK3
	hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
#else
	hbox = gtk_hbox_new (FALSE, 0);
#endif
	gtk_widget_show (hbox);

	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 10);

	/* ========================================================== */
	/* ボード種別 : ラベル */
	label = gtk_label_new ("種別");
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);
	gtk_widget_show(label);

	/* ボード種別 : コンボボックス */
	combo = gtk_combo_box_text_new();
	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "PE");
	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "PCI");

	gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 0);
    gtk_box_pack_start (GTK_BOX (hbox), combo, FALSE, FALSE, 5);
	g_signal_connect (G_OBJECT (combo), "changed",
						G_CALLBACK (bord_kind_combo_box_onchanged), NULL);
	gtk_widget_show(combo);
	/* ---------------------------------------------------------- */

	/* ボードID : ラベル */
	label = gtk_label_new ("ボードID");
    gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 5);
	gtk_widget_show(label);

	/* ボードID : コンボボックス */
	combo = gtk_combo_box_text_new();
	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "0");
	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "1");
	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "2");
	gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "3");

	gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 0);
    gtk_box_pack_start (GTK_BOX (hbox), combo, FALSE, FALSE, 5);
	g_signal_connect (G_OBJECT (combo), "changed",
						G_CALLBACK (bord_id_combo_box_onchanged), NULL);
	gtk_widget_show(combo);
	/* ---------------------------------------------------------- */

	/* フレーム作成 */
	lframe = gtk_frame_new ("");
	gtk_box_pack_start (GTK_BOX (hbox), lframe, FALSE, FALSE, 5);
	gtk_widget_show (lframe);

	/* インフォーメーション表示 */
	g_label_info = gtk_label_new ("");
	gtk_widget_set_size_request (g_label_info, 280, 20);
	gtk_misc_set_alignment (GTK_MISC (g_label_info), 0, 0);
	gtk_container_add (GTK_CONTAINER (lframe), g_label_info);
	gtk_widget_show(g_label_info);

	/* 終了ボタン */
	button = gtk_button_new_with_label ("終了");
	g_signal_connect (G_OBJECT (button), "clicked",
						G_CALLBACK (destroy), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 10);
	gtk_widget_show (button);
	/* ========================================================== */

	/* ========================================================== */
	/* 入出力設定 */
	/* フレーム作成 */
	frame = gtk_frame_new ("入出力設定");
	gtk_container_set_border_width (GTK_CONTAINER (frame), 10);
	gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5);
	gtk_widget_show (frame);

	/* テーブル作成 */
	table = gtk_table_new (4, 7, TRUE);
	gtk_container_set_border_width (GTK_CONTAINER (table), 5);
	gtk_table_set_col_spacings (GTK_TABLE (table), 10);
	gtk_container_add (GTK_CONTAINER (frame), table);
    gtk_widget_show (table);

	/* ---------------------------------------------------------- */
	/* 入力/出力設定ラジオボタン */
	for(i = 0; i < MAX_BIT_NUM; i++)
	{
#ifdef USE_GTK3
		inframe_vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 2);
#else
		inframe_vbox = gtk_vbox_new (FALSE, 2);
#endif
		gtk_table_attach_defaults (GTK_TABLE (table), inframe_vbox, i, i + 1, 0, 3);
		gtk_widget_show (inframe_vbox);

		sprintf(cMsg, "Bit%d", MAX_BIT_NUM - i - 1);
		label = gtk_label_new(cMsg);
		gtk_box_pack_start (GTK_BOX (inframe_vbox), label, FALSE, FALSE, 5);
		gtk_widget_show(label);

		button = gtk_radio_button_new_with_label (NULL, "入力");
		g_signal_connect (G_OBJECT (button), "clicked",
				G_CALLBACK (rc_func[i]),
				(gpointer)DIR_INPUT);
		gtk_box_pack_start (GTK_BOX (inframe_vbox), button, FALSE, FALSE, 5);
		gtk_widget_show (button);
		gDirDataIoRadioBtn[i][DIR_INPUT] = button;

		/* 入力をデフォルト選択 */
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);

		button = gtk_radio_button_new_with_label_from_widget (GTK_RADIO_BUTTON (button), "出力");
		g_signal_connect (G_OBJECT (button), "clicked",
				G_CALLBACK (rc_func[i]),
				DIR_OUTPUT);
		gtk_box_pack_start (GTK_BOX (inframe_vbox), button, FALSE, FALSE, 5);
		gtk_widget_show (button);
		gDirDataIoRadioBtn[i][DIR_OUTPUT] = button;

	}
	/* ---------------------------------------------------------- */

	/* 入力/出力一括設定ボタン */
#ifdef USE_GTK3
	inframe_vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 2);
#else
	inframe_vbox = gtk_vbox_new (FALSE, 2);
#endif
	gtk_table_attach_defaults (GTK_TABLE (table), inframe_vbox, 6, 7, 0, 3);
	gtk_widget_show (inframe_vbox);

	label = gtk_label_new ("");
	gtk_box_pack_start (GTK_BOX (inframe_vbox), label, FALSE, FALSE, 5);
	gtk_widget_show (label);

	button = gtk_button_new_with_label ("全入力");
	g_signal_connect (G_OBJECT (button), "clicked",
						G_CALLBACK (all_set_btn_onclicked),
						(gpointer)DIR_INPUT);
	gtk_box_pack_start (GTK_BOX (inframe_vbox), button, FALSE, FALSE, 5);
	gtk_widget_show (button);

	button = gtk_button_new_with_label ("全出力");
	g_signal_connect (G_OBJECT (button), "clicked",
						G_CALLBACK (all_set_btn_onclicked),
						DIR_OUTPUT);
	gtk_box_pack_start (GTK_BOX (inframe_vbox), button, FALSE, FALSE, 5);
	gtk_widget_show (button);
	/* ---------------------------------------------------------- */

	/* ライトボタン */
	g_entry_write_data = gtk_entry_new();
	gtk_entry_set_max_length (GTK_ENTRY (g_entry_write_data), 2);
	gtk_entry_set_width_chars (GTK_ENTRY (g_entry_write_data), 2);
	gtk_widget_set_size_request (g_entry_write_data, 20, 40);
	gtk_table_attach_defaults (GTK_TABLE (table), g_entry_write_data, 1, 2, 3, 4);
	gtk_widget_show (g_entry_write_data);

	button = gtk_button_new_with_label ("ライト");
	g_signal_connect (G_OBJECT (button), "clicked",
						G_CALLBACK (write_btn_onclicked),
						NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), button, 2, 3, 3, 4);
	gtk_widget_show (button);
	/* ---------------------------------------------------------- */
	/* フレーム作成 */
	lframe = gtk_frame_new ("");
	gtk_table_attach_defaults (GTK_TABLE (table), lframe, 4, 5, 3, 4);
	gtk_widget_show (lframe);

	/* リードデータラベル */
	g_label_read_data = gtk_label_new ("");
	gtk_label_set_width_chars (GTK_LABEL (g_label_read_data), 4);
	gtk_misc_set_alignment (GTK_MISC (g_label_read_data), 0, 0);
	gtk_container_add (GTK_CONTAINER (lframe), g_label_read_data);
	gtk_widget_show(g_label_read_data);

	/* リードボタン */
	button = gtk_button_new_with_label ("リード");
	g_signal_connect (G_OBJECT (button), "clicked",
						G_CALLBACK (read_btn_onclicked), NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), button, 5, 6, 3, 4);
	gtk_widget_show (button);
	/* ========================================================== */

	/* ========================================================== */
	/* 割り込み設定 */
	/* フレーム作成 */
	frame = gtk_frame_new ("割り込み設定");
	gtk_container_set_border_width (GTK_CONTAINER (frame), 10);
	gtk_box_pack_start (GTK_BOX (vbox), frame, FALSE, FALSE, 5);
	gtk_widget_show (frame);

	/* テーブル作成 */
	table = gtk_table_new (4, 7, TRUE);
	gtk_container_set_border_width (GTK_CONTAINER (table), 5);
	gtk_table_set_col_spacings (GTK_TABLE (table), 10);
	gtk_container_add (GTK_CONTAINER (frame), table);
    gtk_widget_show (table);
	/* ---------------------------------------------------------- */

	/* 割り込み閾値設定 */
	for(i = 0; i < MAX_BIT_NUM; i++)
	{
#ifdef USE_GTK3
		inframe_vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 2);
#else
		inframe_vbox = gtk_vbox_new (FALSE, 2);
#endif
		gtk_table_attach_defaults (GTK_TABLE (table), inframe_vbox, i, i + 1, 0, 3);
		gtk_widget_show (inframe_vbox);

		sprintf(cMsg, "Bit%d", MAX_BIT_NUM - i - 1);
		label = gtk_label_new(cMsg);
		gtk_box_pack_start (GTK_BOX (inframe_vbox), label, FALSE, FALSE, 5);
		gtk_widget_show(label);

		/* 閾値 : コンボボックス */
		combo = gtk_combo_box_text_new();
		gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "無し");
		gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "HIGH");
		gtk_combo_box_text_append_text (GTK_COMBO_BOX_TEXT (combo), "LOW");

		gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 0);
		g_signal_connect (G_OBJECT (combo), "changed",
							G_CALLBACK (interrupt_level_combo_box_onchanged),
							(gpointer)(MAX_BIT_NUM - i - 1));
		gtk_box_pack_start (GTK_BOX (inframe_vbox), combo, FALSE, FALSE, 5);
		gtk_widget_show(combo);
	}
	/* ---------------------------------------------------------- */

	/* 割り込みリードデータ */
	label = gtk_label_new("リードデータ");
	gtk_table_attach_defaults (GTK_TABLE (table), label, 2, 4, 3, 4);
	gtk_widget_show(label);

	/* フレーム作成 */
	lframe = gtk_frame_new ("");
	gtk_table_attach_defaults (GTK_TABLE (table), lframe, 4, 5, 3, 4);
	gtk_widget_show (lframe);

	/* 割り込みリードデータラベル */
	g_label_int_data = gtk_label_new ("");
	gtk_label_set_width_chars (GTK_LABEL (g_label_int_data), 4);
	gtk_misc_set_alignment (GTK_MISC (g_label_int_data), 0, 0);
	gtk_container_add (GTK_CONTAINER (lframe), g_label_int_data);
	gtk_widget_show(g_label_int_data);

	/* 割り込み開始ボタン */
	g_button_int = gtk_button_new_with_label ("開始");
	g_signal_connect (G_OBJECT (g_button_int), "clicked",
						G_CALLBACK (interrupt_btn_onclicked), NULL);
	gtk_table_attach_defaults (GTK_TABLE (table), g_button_int, 5, 6, 3, 4);
	gtk_widget_show (g_button_int);
	/* ========================================================== */

	gtk_widget_show(window);

	return;
}

/** 
 * @brief メインルーチン
 * 
 * @param[in] argc		未使用
 * @param[in] argv[]	未使用
 * 
 * @return 0
 */
int main(int argc, char* argv[])
{
	g_type_init();
#if !GLIB_CHECK_VERSION(2, 32, 0)
	g_thread_init(NULL);
#endif
	gtk_init(&argc, &argv);
	gdk_threads_init();

	create_window();

	g_int_flg = eINT_MODE_STOP;
	signal(SIG_DIO_INTERRUPT, interrupt_signal_handler);

	gdk_threads_enter();
	gtk_main();
	gdk_threads_leave();

	driver_close_force();

	signal(SIG_DIO_INTERRUPT, SIG_DFL);

	return 0;
}

